/* ==========================================================================
 * Mandel.js
 *
 * Written by Kostas Symeonidis
 * Copyright (c) 2009-2010 atopon.org, a division of CyLog Software
 * ==========================================================================
 * History:
 *
 * 18.Apr.2010 Converted to use Pixel manipulation (faster by about 40%)
 * 07.Jun.2009 Started experimenting with HTML 5 canvas
 * ==========================================================================
 *
 * ==========================================================================
 * "Mandelbrot Set in HTML 5 Canvas" License
 * GNU General Public License version 3.0 (GPLv3)
 * ==========================================================================
 * This file is part of "Mandelbrot Set in HTML 5 Canvas"
 *
 * "Mandelbrot Set in HTML 5 Canvas" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "Mandelbrot Set in HTML 5 Canvas" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "Mandelbrot Set in HTML 5 Canvas". If not, see <http://www.gnu.org/licenses/>.
 * ==========================================================================
 */

var ctx;
var iCanvasWidth;
var iCanvasHeight;
var iCanvasX;
var iCanvasY;

var timingDiv;

var MAX_CONTROL_COLORS = 5;
var MAX_COLORS = 512;

var controlColors = new Array(MAX_CONTROL_COLORS);
var colors = new Array(MAX_COLORS);

var iterLimit = 10;
var pmin = -2.25;
var pmax = 0.75;
var qmin = -1.5;
var qmax = 1.5;

var lastColor = 0;
var mouseDown = false;
var mbX;
var mbY;

var backImage;
var mandelImage;
var mandelPixels;

function reportCoordsAndTiming ( msg )
{
    document.getElementById("pmin").innerHTML = pmin;
    document.getElementById("pmax").innerHTML = pmax;
    document.getElementById("qmin").innerHTML = qmin;
    document.getElementById("qmax").innerHTML = qmax;
    timingDiv.innerHTML = msg;
}

/**
 * Draw a Pixel on the canvas context. This method is caching the last color used
 * as the ctx.fillStyle is an expensive method.
 *
 * @param x
 * @param y
 * @param c
 */
function drawPixel ( x, y, c )
{
    var iOffset = 4 * (y * iCanvasWidth + x);
    mandelPixels[iOffset    ] = colors[c][0];   // r
    mandelPixels[iOffset + 1] = colors[c][1];   // g
    mandelPixels[iOffset + 2] = colors[c][2];   // b
    mandelPixels[iOffset + 3] = 255;
}

/**
 *
 */
function resetMandel ()
{
    iterLimit = 100;
    qmin = -1.5;
    qmax = 1.5;
    pmin = -2.25;
    pmax = 0.75;
}

/**
 *
 */
function resetControlColors ()
{
    controlColors[0] = [0x00, 0x00, 0x20];
    controlColors[1] = [0xff, 0xff, 0xff];
    controlColors[2] = [0x00, 0x00, 0xa0];
    controlColors[3] = [0x40, 0xff, 0xff];
    controlColors[4] = [0x20, 0x20, 0xff];
}

/**
 *
 */
function computeColors ()
{
    var i, k;
    colors[0] = [0, 0, 0];

    for (i = 0; i < MAX_CONTROL_COLORS - 1; i++)
    {
        var rstep = (controlColors[i + 1][0] - controlColors[i][0]) / 63;
        var bstep = (controlColors[i + 1][1] - controlColors[i][1]) / 63;
        var gstep = (controlColors[i + 1][2] - controlColors[i][2]) / 63;

        for (k = 0; k < 64; k++)
        {
            colors[k + (i * 64) + 1] = [Math.round(controlColors[i][0] + rstep * k),
                                        Math.round(controlColors[i][1] + bstep * k),
                                        Math.round(controlColors[i][2] + gstep * k)];
        }
    }

    for (i = 257; i < MAX_COLORS; i++)
    {
        colors[i] = colors[i - 256];
    }
}

function computeMandel ()
{
    var kmax = 4096;
    var xstep = (pmax - pmin) / iCanvasWidth;
    var ystep = (qmax - qmin) / iCanvasHeight;

    // declare and initialise variables, for speed
    var x = 0.0;
    var y = 0.0;
    var r = 1.0;

    // create a back image and get a pointer to the pixels array
    mandelImage = ctx.getImageData(0, 0, iCanvasWidth, iCanvasHeight);
    mandelPixels = mandelImage.data;

    var start = new Date().getTime();
    var a = 2 * 0.2 * Math.PI, sc=0.125, s = sc*Math.sin(a), c = sc*Math.cos(a);
    for (var sx = 0; sx < iCanvasWidth; sx++)
    {
        for (var sy = 0; sy < iCanvasHeight; sy++)
        {
            var p = pmin + xstep * sx;
            var q = qmax - ystep * sy;
            var k = 0;
            //var x0 = c*p + s*q + 1;
            //var y0 = -s * p + c * q + 1;
            var x0 = 0, y0 = 0;

            do
            {
                x = x0 * x0 - y0 * y0 + p;
                y = 2 * x0 * y0 + q;
                x0 = x;
                y0 = y;
                r = x * x + y * y;
                k++;
            }
            while ((r <= iterLimit) && (k < kmax));

            if (k >= kmax)
            {
                k = 0;
            }

            // draw the pixel
            drawPixel(sx, sy, k % 256);
        }

    }

    ctx.putImageData(mandelImage, 0, 0);

    var elapsed = new Date().getTime() - start;
    reportCoordsAndTiming(elapsed + " milliseconds");
}

function onMouseDown ( e )
{
    e = window.event || e;
    mouseDown = true;

    mbX = e.offsetX || (e.clientX - iCanvasX);
    mbY = e.offsetY || (e.clientY - iCanvasY);

    backImage = ctx.getImageData(0, 0, iCanvasWidth, iCanvasHeight);
}

function onMouseMove ( e )
{
    e = window.event || e;
    if (mouseDown)
    {
        var currX = e.offsetX || (e.clientX - iCanvasX);
        var currY = e.offsetY || (e.clientY - iCanvasY);
        var newY = mbY + (booleanToInt(currY > mbY) * 2 - 1) * Math.round(iCanvasHeight * Math.abs(currX - mbX) / iCanvasWidth);

        ctx.putImageData(backImage, 0, 0);
        ctx.strokeStyle = "rgb(170,255,65)";
        ctx.strokeRect(mbX, mbY, currX - mbX, newY - mbY);
    }

}

function onMouseUp ( e )
{
    e = window.event || e;
    if (mouseDown)
    {
        var currX = e.offsetX || (e.clientX - iCanvasX);
        var currY = e.offsetY || (e.clientY - iCanvasY);

        var newX = currX;
        var newY = mbY + (booleanToInt(currY > mbY) * 2 - 1) * Math.round(iCanvasHeight * Math.abs(currX - mbX) / iCanvasWidth);

        log(mbX + ', ' + mbY + ' : ' + newX + ', ' + newY);

        if (newX < mbX)
        {
            var hx = newX;
            newX = mbX;
            mbX = hx;
        }
        if (newY < mbY)
        {
            var hy = newY;
            newY = mbY;
            mbY = hy;
        }

        log(mbX + ', ' + mbY + ' ' + newX + ', ' + newY);

        // only bother if the size of the square is more than 3x3 pixels
        if ((Math.abs(newX - mbX) > 3) && (Math.abs(newY - mbY) > 3))
        {
            var pw = pmax - pmin;
            pmin = pmin + mbX * pw / iCanvasWidth;
            pmax = pmax - (iCanvasWidth - newX) * pw / iCanvasWidth;
            var qw = qmax - qmin;
            qmin = qmin + (iCanvasHeight - newY) * qw / iCanvasHeight;
            qmax = qmax - mbY * qw / iCanvasHeight;

            computeMandel();
        }
    }
    mouseDown = false;
}


/**
 *
 * @param canvasElement
 * @param w
 * @param h
 */
function initMandel ( canvasElement, w, h )
{
    iCanvasWidth = w;
    iCanvasHeight = h;
    var canvas = document.getElementById(canvasElement);
    ctx = canvas.getContext("2d");
    canvas.width = w;
    canvas.height = h;

    // get the location of the canvas on the page
    iCanvasX = canvas.offsetLeft;
    iCanvasY = canvas.offsetTop;

    resetMandel();
    resetControlColors();
    computeColors();

    canvas.onmousedown = onMouseDown;
    canvas.onmousemove = onMouseMove;
    canvas.onmouseup = onMouseUp;

    timingDiv = document.getElementById("timing");

    log('+ canvas initialised at ' + iCanvasX + ' ' + iCanvasY);
}